#!/usr/bin/perl -w

# Copyright 2014, 2015, 2016, 2017, 2018, 2019, 2021, 2022 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use Math::Geometry::Planar ();
use Math::PlanePath::CCurve 117; # v.117 for level_to_n_range()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyLineSegments;
use MyCCurve 'chamfer','xy_rotate_minus90';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 9;
my $nosides = 0;

if (@ARGV && $ARGV[0] eq 'nosides') {
  $nosides = 1;
}

print "  % generated by tools/$FindBin::Script",
  ($nosides ? " nosides" : ""),"\n";
print "  \\newcommand\\MyLevel{$level}\n";
my $path = Math::PlanePath::CCurve->new;

my ($n_lo, $n_hi) = $path->level_to_n_range($level);
my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

sub transform {
  my ($x,$y) = @_;
  MyCCurve::transform_to_horizontal($x,$y, $level);
}

my $segments = MyLineSegments->new;   # not transformed yet
foreach my $n ($n_lo .. $n_hi-1) {
  $segments->add($path->n_to_xy($n),
                 $path->n_to_xy($n+1));
}

# convex hull points
my $points = [];
foreach my $line ($segments->lines) {
  my ($x1,$y1, $x2,$y2) = @$line;
  push @$points, [$x1,$y1], [$x2,$y2];
}
my $planar = Math::Geometry::Planar->new;
$planar->points($points);
$planar = $planar->convexhull2;
$points = $planar->points;
@$points = sort {
  my ($xa,$ya) = transform(@$a);
  my ($xb,$yb) = transform(@$b);
  ($yb <=> $ya)        # descending Y
    || ($xb <=> $xa)   # descending X
  } @$points;

# Remove segments which are overwritten by diagonal minimum area rectangle
# sides.
if ($nosides) {
  my $points = MyCCurve::hull_points_sort_by_angle($points, $x_hi,$y_hi);
  foreach my $i (0 .. $#$points) {
    my ($x1,$y1) = @{$points->[$i-1]};
    my ($x2,$y2) = @{$points->[$i]};
    my $dx = $x2 - $x1;
    my $dy = $y2 - $y1;
    ### dxdy: "$dx,$dy"
    if ($dx==0 || $dy==0) {
      ### remove: "$x1,$y1, $x2,$y2"
      $segments->remove($x1,$y1, $x2,$y2);
    }
  }
}


$Printwrap = 2;
printwrap("\\draw[my grey] ");
$Printwrap = 4;
$segments->printwrap (transform => \&transform);
printwrap(";\n");

$Printwrap = 2;
{
  my ($x,$y) = transform(0,0);
  printwrap("\\coordinate (start) at ($x,$y);\n");
}
{
  my ($x,$y) = transform($x_hi,$y_hi);
  printwrap("\\coordinate (end) at ($x,$y);\n");
}

my $num = 1;
my $dash = '';
foreach my $p (@$points) {
  my $name = "P$num$dash";
  my ($x,$y) = @$p;
  print "% GP-Test-Last  $name($level) == $x + I*$y\n";
  ($x,$y) = transform($x,$y);
  print "  \\node ($name) at ($x,$y) [my dot];\n";

  if ($dash) {
    $num++;
    $dash = '';
  } else {
    $dash = 'dash';
  }
}

print "  % end generated\n";
exit 0;
